# -*- coding: utf-8 -*-
"""fd-vpn.classes.modules"""

import abc
import logging
import os
import time
import base64
import shlex
import shutil
import subprocess
import tempfile
import zipfile

import qrcode
import pyotp

from six import BytesIO, iteritems, itervalues

from OpenSSL import crypto

from dwho.classes.modules import DWhoModuleWebBase
from dwho.classes.plugins import PLUGINS
from sonicprobe import helpers
from sonicprobe.libs.gencert import GenCert
from httpdis.ext.httpdis_json import HttpResponse, HttpReqErrJson

LOG              = logging.getLogger('fd-vpn.classes.modules')
OPENVPN_EXE_NAME = 'openvpn-installer.exe'

class VPNModuleClientBase(DWhoModuleWebBase):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def MODULE_NAME(self):
        return

    @abc.abstractproperty
    def LOCK(self):
        return

    def __init__(self):
        self.objects = {}

        DWhoModuleWebBase.__init__(self)

    def init(self, config):
        DWhoModuleWebBase.init(self, config)

        return self

    def safe_init(self, options):
        self.lock_timeout       = self.config['general']['lock_timeout']

        self.base_path          = self.config['vpn']['client']['base_path']
        self.ca_name            = self.config['vpn']['client']['ca_name']
        self.common_name        = self.config['vpn']['client']['common_name']

        self.ca_path            = os.path.join(self.base_path, 'ca')
        self.client_path        = os.path.join(self.base_path, 'client')

        self.qrcode_issuer_name = self.config['vpn']['client']['qrcode_issuer_name']

        self.security_link      = self.config['vpn']['client']['security_link']

        self.plugin_iptables    = PLUGINS['iptables']
        self.plugin_openvpn     = PLUGINS['openvpn']
        self.plugin_credentials = PLUGINS['credentials']

    def _fetch(self, xid, secure = True):
        r = self.objects['vpn_client'].get(to_dict = True,
                                           **{'id': xid})
        if secure:
            r.pop('auth_data', None)

        return r

    def _fetch_by_reference(self, reference, secure = True):
        r = self.objects['vpn_client'].get(to_dict = True,
                                           **{'reference': reference})
        if r and secure:
            r.pop('auth_data', None)

        return r

    def _fetch_by_security_reference(self, security_reference, secure = True, expired = False):
        r = self.objects['vpn_client'].get(to_dict = True,
                                           **{'security_reference': security_reference})
        if not r:
            return r

        if not expired \
           and time.strftime('%Y-%m-%d %H:%M:%S') > r['security_expired_at']:
            return False

        if secure:
            r.pop('auth_data', None)

        return r

    def _generate_installer(self, server, client, arch, files):
        build_dir = None
        cwd       = os.getcwd()

        try:
            build_dir   = tempfile.mkdtemp()
            os.chdir(build_dir)

            confvpn_dir = os.path.join(build_dir, 'OpenVPN', 'config')

            config_dir  = os.path.join(confvpn_dir, server['reference'])
            os.makedirs(config_dir)

            config_file = os.path.join(confvpn_dir, "%s.ovpn" % client['reference'])
            config_inline_file = os.path.join(confvpn_dir, "%s-inline.ovpn" % client['reference'])

            """
            helpers.file_w_tmp(self.plugin_openvpn.get_client_config(client['reference'],
                                                                     server).getvalue(),
                               config_file)
            """

            with open(config_file, 'w+') as f:
                f.write(self.plugin_openvpn.get_client_config(client['reference'],
                                                              client['auth_method'],
                                                              server).getvalue())

            with open(config_inline_file, 'w+') as f:
                f.write(self.plugin_openvpn.get_client_config_inline(client['reference'],
                                                                     client['auth_method'],
                                                                     server,
                                                                     files).getvalue())

            for xfile in itervalues(files):
                shutil.copy2(xfile, config_dir)

            if arch == 'ovpn':
                os.chdir(confvpn_dir)
                outfile = os.path.join(build_dir, self.config['vpn']['client']['zip_outfile'])
                zipf = zipfile.ZipFile(outfile, 'w', zipfile.ZIP_DEFLATED)
                for wroot, wdirs, wfiles in os.walk('.'): # pylint: disable-msg=unused-variable
                    for wfile in wfiles:
                        zipf.write(os.path.join(wroot, wfile))
                zipf.close()

                with open(outfile, 'rb') as f:
                    outdata = f.read()

                return HttpResponse(data    = outdata,
                                    headers = {'Content-type': 'application/zip',
                                               'Content-Disposition': 'inline; filename="%s-%s-%s.zip"'
                                                                       % (client['reference'],
                                                                          arch,
                                                                          self.config['vpn']['client']['zip_outfile'])})

            exe_file    = self.config['vpn']['client']['openvpn_exe'][arch]

            shutil.copyfile(exe_file, os.path.join(build_dir, OPENVPN_EXE_NAME))

            cmd         = subprocess.Popen(shlex.split(self.config['vpn']['client']['nsi_cmd']
                                                       % {'build_dir':        build_dir,
                                                          'client_reference': client['reference'],
                                                          'server_reference': server['reference'],
                                                          'cidr':             server['cidr']}),
                                           stderr       = subprocess.PIPE,
                                           close_fds    = True)
            (out, err)  = cmd.communicate() # pylint: disable-msg=unused-variable

            if cmd.returncode:
                raise RuntimeError("unable to generate installer. (build_dir: %r, err: %r, rc: %r)"
                                   % (build_dir, err, cmd.returncode))

            outfile     = os.path.join(build_dir, self.config['vpn']['client']['nsi_outfile'])

            with open(outfile, 'rb') as f:
                outdata = f.read()

            return HttpResponse(data    = outdata,
                                headers = {'Content-type': 'application/octet-stream',
                                           'Content-Disposition': 'inline; filename="%s-%s-%s"'
                                                                   % (client['reference'],
                                                                      arch,
                                                                      self.config['vpn']['client']['nsi_outfile'])})
        finally:
            if os.path.isdir(cwd):
                os.chdir(cwd)
            if build_dir:
                shutil.rmtree(build_dir)

    def _get(self, server, client):
        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            client['groupname']   = None
            client['group_id']    = None
            client['group_color'] = None
            client['auth_qrcode'] = None

            if server['group_id']:
                group = self.objects['vpn_group'].get(**{'id': server['group_id']})
                if group:
                    client['groupname']   = group['groupname']
                    client['group_id']    = group['id']
                    client['group_color'] = group['color']

            client['ip_addr']  = '.'.join(server['cidr'].split('.')[0:3] + [client['last_octet']])

            if client['auth_method'] == 'otp':
                secret_url = pyotp.totp.TOTP(client['auth_data']).provisioning_uri(
                    name=client['reference'],
                    issuer_name=self.qrcode_issuer_name)

                buffered = BytesIO()
                img = qrcode.make(secret_url)
                img.save(buffered, format = 'PNG')
                client['auth_qrcode'] = base64.b64encode(buffered.getvalue())

            client.pop('auth_data', None)

            return client
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _download(self, server, client, arch):
        if not client:
            raise HttpReqErrJson(404, "unable to find client. (id: %r)" % client['id'])

        if not server:
            raise HttpReqErrJson(404, "unable to find server. (id: %r)" % client['server_id'])

        if not server['group_id']:
            raise HttpReqErrJson(415, "invalid server. (id: %r)" % server['id'])

        obj             = {}
        obj.update(dict(["vpn_client.%s" % x, y] for x, y in iteritems(client)))
        obj.update(dict(["vpn_server.%s" % x, y] for x, y in iteritems(server)))

        ca_crt_path     = os.path.join(self.ca_path, "%s.crt" % self.ca_name)
        ca_pkey_path    = os.path.join(self.ca_path, "%s.key" % self.ca_name)
        ca_srl_path     = os.path.join(self.ca_path, "%s.srl" % self.ca_name)

        if not os.access(ca_crt_path, (os.R_OK | os.F_OK)):
            raise HttpReqErrJson(400, "unable to find the CA certificate: %r" % ca_crt_path)

        if not os.access(ca_pkey_path, (os.R_OK | os.F_OK)):
            raise HttpReqErrJson(400, "unable to find the PKey certificate: %r" % ca_pkey_path)

        if not os.access(ca_srl_path, (os.R_OK | os.F_OK)):
            raise HttpReqErrJson(400, "unable to find the serial number: %r" % ca_srl_path)

        attributes                  = self.config['vpn']['client']['attributes'].copy()
        attributes['commonName']    = self.common_name % obj

        client_pkey_path            = os.path.join(self.client_path, "%s.key" % attributes['commonName'])
        client_crt_path             = os.path.join(self.client_path, "%s.crt" % attributes['commonName'])

        config                      = self.plugin_openvpn.get_server_config(server['reference'])[1]

        files                       = {'ca':        ca_crt_path,
                                       'pkey':      client_pkey_path,
                                       'crt':       client_crt_path,
                                       'ta_key':    config.get('general', 'tls-auth').split(' ', 1)[0]}

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        error           = False
        client_csr_path = None

        try:
            if os.access(client_pkey_path, (os.R_OK | os.F_OK)) \
               and os.access(client_crt_path, (os.R_OK | os.F_OK)):
                with open(client_pkey_path, 'r') as f:
                    client_pkey = crypto.load_privatekey(crypto.FILETYPE_PEM,
                                                         ''.join(f.readlines()))

                with open(client_crt_path, 'r') as f:
                    client_crt  = crypto.load_certificate(crypto.FILETYPE_PEM,
                                                          ''.join(f.readlines()))

                if not client_crt.has_expired():
                    crypto.dump_privatekey(crypto.FILETYPE_PEM, client_pkey)
                    crypto.dump_certificate(crypto.FILETYPE_PEM, client_crt)

                    return self._generate_installer(server,
                                                    client,
                                                    arch,
                                                    files)

            LOG.info("!!!! %r !!!", self.client_path)
            client_csr_path = os.path.join(self.client_path, "%s.csr" % attributes['commonName'])

            with open(ca_crt_path, 'r') as f:
                ca_crt  = crypto.load_certificate(crypto.FILETYPE_PEM,
                                                  ''.join(f.readlines()))

            with open(ca_pkey_path, 'r') as f:
                ca_pkey = crypto.load_privatekey(crypto.FILETYPE_PEM,
                                                 ''.join(f.readlines()))

            is_hex = False

            with open(ca_srl_path, 'r') as f:
                x = f.readline().strip()
                is_hex = not x.isdigit()
                if is_hex:
                    ca_srl = int(x, 16) + 1
                else:
                    ca_srl = int(x) + 1

            gen_cert        = GenCert(self.config['vpn']['client']['options']['bits'],
                                      self.config['vpn']['client']['options']['crypto'],
                                      self.config['vpn']['client']['options']['digest'],
                                      notafter_days = self.config['vpn']['client']['options']['days'])

            pkey            = gen_cert.make_privatekey(client_pkey_path)
            csr             = gen_cert.make_certreq(pkey,
                                                    attributes,
                                                    client_csr_path,
                                                    keyUsage = ['digitalSignature', 'keyAgreement'],
                                                    extendedKeyUsage = ['clientAuth'])

            crt             = gen_cert.make_certificate(csr,
                                                        ca_crt,
                                                        ca_pkey,
                                                        ca_srl,
                                                        client_crt_path)

            crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey)
            crypto.dump_certificate(crypto.FILETYPE_PEM, crt)

            if is_hex:
                srl = format(crt.get_serial_number(), 'x').upper()
            else:
                srl = crt.get_serial_number()

            helpers.file_w_tmp("%s" % srl, ca_srl_path, mode = 'w+')

            return self._generate_installer(server,
                                            client,
                                            arch,
                                            files)
        except HttpReqErrJson:
            error = True
            raise
        except Exception as e:
            error = True
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if error:
                if os.path.exists(client_crt_path):
                    os.unlink(client_crt_path)
                if os.path.exists(client_pkey_path):
                    os.unlink(client_pkey_path)
                if os.path.exists(client_csr_path):
                    os.unlink(client_csr_path)
            self.LOCK.release()
