# -*- coding: utf-8 -*-
"""fd-vpn.modules.vpn_enduser"""

import logging
import os
import re
import time
import uuid

import bcrypt
import pyotp

from uuid import uuid4
from datetime import datetime, timedelta

from six import ensure_str, iteritems

from fd_vpn.classes.modules import VPNModuleClientBase
from fd_vpn.objects.vpn_client import FdVpnObjectVPNClient
from fd_vpn.objects.vpn_server import FdVpnObjectVPNServer
from fd_vpn.objects.vpn_group import FdVpnObjectVPNGroup

from dwho.classes.modules import DWhoModuleSQLBase, MODULES
from sonicprobe import helpers
from sonicprobe.libs import xys
from sonicprobe.libs.moresynchro import RWLock
from sonicprobe.validator.email import SPValidatorEmail
from httpdis.ext.httpdis_json import HttpResponse, HttpReqError, HttpReqErrJson


LOG = logging.getLogger('fd-vpn.modules.vpn_enduser')

xys.add_regex('vpn_enduser.reference', re.compile(r'^[a-zA-Z0-9]{8}$').match)
xys.add_regex('vpn_enduser.security_reference', re.compile(r'^[0-9a-f]{8}\-[0-9a-f]{4}\-4[0-9a-f]{3}\-[89ab][0-9a-f]{3}\-[0-9a-f]{12}$').match)
xys.add_callback('vpn_enduser.password_check', helpers.password_check)


class VPNEndUserModule(DWhoModuleSQLBase, VPNModuleClientBase):
    MODULE_NAME = 'vpn_enduser'

    LOCK        = RWLock()

    def __init__(self):
        DWhoModuleSQLBase.__init__(self)
        VPNModuleClientBase.__init__(self)

        self.objects = {'vpn_client': FdVpnObjectVPNClient(),
                        'vpn_server': FdVpnObjectVPNServer(),
                        'vpn_group':  FdVpnObjectVPNGroup()}

    def init(self, config):
        DWhoModuleSQLBase.init(self, config)
        VPNModuleClientBase.init(self, config)

        self.objects['vpn_client'].init(config)
        self.objects['vpn_server'].init(config)
        self.objects['vpn_group'].init(config)


    GET_SCHEMA  = xys.load("""
    security_reference: !~~regex(vpn_enduser.security_reference)
    """)

    def get(self, request):
        options = request.query_params()
        if not xys.validate(options, self.GET_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        client = self._fetch_by_security_reference(options['security_reference'], secure = False)
        if not client:
            raise HttpReqErrJson(404, "client not found. (security_reference: %r)" % options['security_reference'])

        server = self.objects['vpn_server'].get(**{'id': client['server_id']})
        if not server:
            raise HttpReqErrJson(404, "server not found. (id: %r)" % client['server_id'])

        return self._get(server, client)


    DOWNLOAD_SCHEMA  = xys.load("""
    security_reference: !~~regex(vpn_enduser.security_reference)
    arch:               !~~enum(x32,x64,xp_x32,xp_x64,ovpn)
    """)

    def download(self, request):
        options = request.query_params()
        if not xys.validate(options, self.DOWNLOAD_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        client  = self._fetch_by_security_reference(options['security_reference'])
        if not client:
            raise HttpReqErrJson(404, "unable to find client. (security_reference: %r)" % options['security_reference'])

        server  = self.objects['vpn_server'].get(to_dict = True,
                                                 **{'id': client['server_id']})
        if not server:
            raise HttpReqErrJson(404, "unable to find server. (id: %r)" % client['server_id'])

        return self._download(server, client, options['arch'])


    CHANGE_AUTH_DATA_SCHEMA   = xys.load("""
    security_reference-: !~~regex(vpn_enduser.security_reference)
    auth_data-: !!str
    """)

    def change_auth_data(self, request):
        args = request.payload_params()
        if not xys.validate(args, self.CHANGE_AUTH_DATA_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        client = self._fetch_by_security_reference(args['security_reference'])
        if not client:
            raise HttpReqErrJson(404, "unable to find client. (security_reference: %r)" % options['security_reference'])

        if client['auth_method'] != 'userpass':
            raise HttpReqErrJson(415, "invalid authentification method, must be 'userpass'")

        server = self.objects['vpn_server'].get(**{'id': client['server_id']})
        if not server:
            raise HttpReqErrJson(404, "unable to find server. (id: %r)" % args['server_id'])

        if not server['group_id']:
            raise HttpReqErrJson(415, "invalid server. (id: %r)" % server['id'])

        try:
            try:
                helpers.password_check(args['auth_data'])
            except Exception as e:
                raise HttpReqErrJson(415, repr(e))

            data = {'auth_data': bcrypt.hashpw(ensure_str(args['auth_data']),
                                               bcrypt.gensalt())}

            return self.objects['vpn_client'].update(data.keys(),
                                                     data.values(),
                                                     **{'security_reference': args['security_reference']})
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    VALIDATE_SCHEMA = xys.load("""
    auth_data-: !~~callback(vpn_client.password_check)
    """)

    def validate(self, request):
        options = request.query_params()
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            for key, value in iteritems(options):
                if not self.VALIDATE_SCHEMA.has_key(key):
                    raise HttpReqErrJson(404, "unknown argument for command. (key: %r)" % key)

                if not xys.validate({key: value}, {key: self.VALIDATE_SCHEMA[key]}):
                    raise HttpReqErrJson(415, "invalid argument for command. (key: %r)" % key)

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(VPNEndUserModule())
    _start()
